#! /usr/bin/perl -w

use strict;
use Net::IRC;
use XINI;
use POSIX ':sys_wait_h';
use Net::hostent;
use MIME::Base64;

my $VERSION = '3.0';

if (@ARGV and $ARGV[0] eq '--version') {
    print "Tuxaator $VERSION\n";
    exit;
}

our @CONFIGFILES = @ARGV;

our %SERVICES; # list of all currently active services
our @CHANNELS; # list of all currently active channels
our %RULES;
# our %LEVEL_CACHE; # userhost => [level, timestamp]
our @user_patterns; # list of [pattern => level] pairs
our $reinitialisation_pending = 0;
our $processing_event = 0;

sub initialise ();
sub load_user_patterns ($);
sub userlevel ($);
sub atexpand ($%);

init_ini @CONFIGFILES;

my $irc = new Net::IRC;
my $conn = $irc->newconn(Nick => ini 'general:nick',
                         Server => ini 'general:host',
                         Port => ini 'general:port = 6667',
                         Ircname => ini "general:ircname = Tuxaator 3.0");

initialise;

{
    package Spawn;

    use strict;
    use IPC::Open2;
    use Fcntl qw(F_GETFL F_SETFL O_NONBLOCK);

    sub new ($$$) {
        my ($class, $command, $output_handler) = @_;
        my $this = {};

        $this->{command} = $command;
        $this->{output_handler} = $output_handler;

        my ($rdh, $wrh);
        my $pid = open2($rdh, $wrh, $command);
        my $flags = fcntl $rdh, F_GETFL, 0 or die "fcntl: F_GETFL: $!";
        fcntl $rdh, F_SETFL, $flags | O_NONBLOCK or die "fcntl: F_SETFL: $!";

        $this->{pid} = $pid;
        $this->{stdout} = $rdh;
        $this->{stdin} = $wrh;
        $this->{input_end} = 0;
        $this->{in_queue} = '';
        $this->{out_queue} = '';
        $this->{selecting_input} = 0;

        $irc->addfh($this->{stdout}, sub {$this->process_output}, 'r');

        return bless $this, $class;
    }

    sub process_input ($) {
        my ($this) = @_;

        return unless defined $this->{stdin};

        my $result = 1;
        while ($result) {
            last if $this->{in_queue} eq '';
            $result = syswrite($this->{stdin}, $this->{in_queue},
                               length $this->{in_queue}, 0);
            goto CLOSE_DOWN if $!{EPIPE};
            die "write to service: $!\n"
                unless defined $result or $!{EAGAIN};
            substr($this->{in_queue}, 0, $result) = ''
                if defined $result;
        }
        if ($this->{in_queue} ne '' and !$this->{selecting_input}) {
            $irc->addfh($this->{stdin}, sub {$this->process_input}, 'w');
            $this->{selecting_input} = 1;
        }
        if ($this->{in_queue} eq '' and $this->{selecting_input}) {
            $irc->removefh($this->{stdin});
            $this->{selecting_input} = 0;
        }
        if ($this->{input_end} and $this->{in_queue} eq '') {
CLOSE_DOWN:
            $irc->removefh($this->{stdin}) if $this->{selecting_input};
            close $this->{stdin};
            $this->{stdin} = undef;
            $this->{in_queue} = undef;
        }
        return;
    }

    sub process_output ($) {
        my ($this) = @_;

        my $result;
        do {
            $result = sysread($this->{stdout}, $this->{out_queue},
                              1024, length $this->{out_queue});
            if (!defined $result and !$!{EAGAIN}) {
	        my @errors = grep {$!{$_}} sort keys(%!);
                warn "service \"$this->{command}\" (pid $this->{pid}) died anomalously (\@errors = (@errors); \$! = \"$!\")";
                goto CLEANUP;
            }
        } while ($result);
        if (defined $result and $result == 0) {
CLEANUP:
            $irc->removefh($this->{stdout});
            close $this->{stdout};
            $this->{stdout} = undef;
            $this->{out_queue} =~ s/([^\n])$/$1\n/s;
        }
        my $interpreter = $this->{output_interpreter};
        while ($this->{out_queue} =~ s/^(.*?)\r?\n//s) {
            &{$this->{output_handler}}($1);
        }
        return;
    }

    sub write ($$) {
        my ($this, $data) = @_;

        $this->{in_queue} .= $data if defined $this->{stdin};
	$this->process_input;
        return;
    }

    sub close_input ($) {
        my ($this) = @_;

        $this->{input_end} = 1;
        $this->process_input;
        return;
    }
}

{
    package Service;

    use strict;
    use IPC::Open2;
    use Fcntl qw(F_GETFL F_SETFL O_NONBLOCK);
    use MIME::Base64;
    use XINI;

    sub new ($$) {
        my ($class, $name) = @_;
        my $this = {};

        $this->{name} = $name;
        my $type = lc ini "service $name:type=daemon";
        if ($type eq 'daemon') {
            $this->{spawn} = new Spawn ini "service $name:command", sub {
                my ($line) = @_;
                my $interpreter = $this->{output_interpreter};
                $this->$interpreter($line);
            };
        } elsif ($type eq 'invoke') {
            $this->{command} = ini "service $name:command";
        } else {
            die "unknown service type '$type' for '$name'";
        }
        $this->{type} = $type;

        my $format_input = ini "service $name:format input=complex";
        if ($format_input =~ m/^complex$/i) {
            $this->{input_formatter} = 'format_complex';
        } elsif ($format_input =~ m/^chat$/i) {
            $this->{input_formatter} = 'format_chat';
        } elsif ($format_input =~ m/^table:\s*(.*)$/i) {
            $this->{input_fields} = [map {"event_field_$_"} split /\s+/, lc $1];
            my %known_fields = map {("event_field_$_" => 1)} qw(type from to nick userhost args time gmtime localtime context);
            for my $field (@{$this->{input_fields}}) {
                die "unknown field '$field' in [service $name]format input"
                    unless exists $known_fields{$field};
            }
            $this->{input_formatter} = 'format_tabs';
            my $separator;
            if (ini_got "service $name:column separator") {
                $separator = ini "service $name:column separator";
                $separator = ' ' if $separator eq '';
            } else {
                $separator = "\t";
            }
            $this->{column_separator} = $separator;
        } else {
            die "unknown value [service $name]format input = $format_input";
        }

        my $interpret_output = ini "service $name:interpret output=stateless commands";
        if ($interpret_output =~ m/^stateless\s+commands$/i) {
            $this->{output_interpreter} = 'stateless_commands';
	} elsif ($interpret_output =~ m/^relay$/i) {
            $this->{output_interpreter} = 'relay';
        } else {
            die "unknown value [service $name]interpret output = $interpret_output";
        }

        return bless $this, $class;
    }

    sub enter ($$) {
        my ($this, $event) = @_;

        if ($this->{main_rule}->matches($event)) {
	    if ($this->{main_rule}->honourable($event)) {
                $this->handle($event);
	    } else {
	        &main::insult($event);
	    }
	}
        return;
    }

    sub handle ($$) {
        my ($this, $event) = @_;

        my $format = $this->{input_formatter};
        my $data = $this->$format($event);

        if ($this->{type} eq 'daemon') {
            $this->{spawn}->write("$data\n");
	    print "wrote \"$data\" into the daemon\n";
        } elsif ($this->{type} eq 'invoke') {
            my $context = $this->event_field_context($event);
	    my %subst = (
                TYPE => sub { $this->event_field_type($event) },
                FROM => sub { $this->event_field_from($event) },
                TO => sub { $this->event_field_to($event) },
                NICK => sub { $this->event_field_nick($event) },
                USERHOST => sub { $this->event_field_userhost($event) },
                ARGS => sub { join '', $this->event_field_args($event) },
                REQARGS => sub { $this->event_field_reqargs($event) },
                TIME => sub { $this->event_field_time($event) },
                LOCALTIME => sub { $this->event_field_localtime($event) },
                GMTIME => sub { $this->event_field_gmtime($event) },
                CONTEXT => sub { $this->event_field_context($event) },
                LEVEL => sub { $this->event_field_level($event) },
		PID => $$,
		CONFIGFILES => join(' ', @main::CONFIGFILES),
	    );
            my $command = $this->{command};
            $command = &main::atexpand($command, %subst);
            my $spawn = new Spawn $command, sub {
                my ($line) = @_;
                $conn->privmsg($context, $line);
            };
            $spawn->write("$data\n");
            $spawn->close_input;
        }
        return;
    }

    sub format_complex ($$) {
        my ($this, $event) = @_;

        my $line = $event->type;
        $line .= " from:" . encode_base64($event->from, "");
        $line .= " to:" . join(',', map {encode_base64($_, "")} $event->to);
        $line .= " nick:" . encode_base64($event->nick, "");
        $line .= " userhost:" . encode_base64($event->userhost, "");
        $line .= " args:" . join(',', map {encode_base64($_, "")} $event->args);
        return $line;
    }

    sub format_chat ($$) {
        my ($this, $event) = @_;

        my $line = '';
	if ($event->type eq 'public') {
	    $line .= "<" . $event->nick . ">";
	} elsif ($event->type eq 'caction') {
	    $line .= "* " . $event->nick;
	} elsif ($event->type eq 'mode') {
	    $line .= "@ " . $event->nick;
	} elsif ($event->type eq 'join') {
	    $line .= "Joining: " . $event->nick;
	} elsif ($event->type eq 'part') {
	    $line .= "Parting: " . $event->nick;
	} elsif ($event->type eq 'quit') {
	    $line .= "Quitting: " . $event->nick;
	} elsif ($event->type eq 'quit') {
	    $line .= "Renicking: " . $event->nick;
	} else {
	    $line .= $event->type . " [" . $event->nick . "]";
	}
        $line .= " " . join(' ', $event->args);
        return $line;
    }

    sub event_field_type ($$) {
        my ($this, $event) = @_;

        return $event->type;
    }

    sub event_field_from ($$) {
        my ($this, $event) = @_;

        return $event->from;
    }

    sub event_field_to ($$) {
        my ($this, $event) = @_;

        return $event->to;
    }

    sub event_field_nick ($$) {
        my ($this, $event) = @_;

        return $event->nick;
    }

    sub event_field_userhost ($$) {
        my ($this, $event) = @_;

        return $event->userhost;
    }

    sub event_field_args ($$) {
        my ($this, $event) = @_;

        return $event->args;
    }

    sub event_field_reqargs ($$) {
        my ($this, $event) = @_;

        my $args = join ' ', $event->args;
	$args =~ s/^\S+\s*//;
	return $args;
    }

    sub event_field_time ($$) {
        return time;
    }

    sub event_field_localtime ($$) {
        return scalar localtime;
    }

    sub event_field_gmtime ($$) {
        return scalar gmtime;
    }

    sub event_field_context ($$) {
        my ($this, $event) = @_;

        my $type = $event->type;
        if ($type eq 'public') {
            return join(",", $event->to);
        } else {
            return $event->nick;
        }
    }

    sub event_field_level ($$) {
        my ($this, $event) = @_;

        return &main::userlevel($event->userhost);
    }

    sub format_tabs ($$) {
        my ($this, $event) = @_;
        my @items;
        for my $field (@{$this->{input_fields}}) {
            my @data = $this->$field($event);
            push @items, join(' ', @data);
        }
        return join $this->{column_separator}, @items;
    }

    sub stateless_commands ($$) {
        my ($this, $command) = @_;
        if ($command =~ m/^\/say\s+(.*)$/si) {
            for my $channel (@CHANNELS) {
                $conn->privmsg($channel, $1);
            }
        } elsif ($command =~ m/^\/msg\s+(\S+)\s+(.*)$/si) {
            my $msg = $2;
            for my $dest (split /,/, $1) {
                $conn->privmsg($dest, $msg);
            }
        } else {
            warn "Service $this->{name} issued weird stateless command \"$command\".\n";
        }
    }

    sub relay ($$) {
        my ($this, $command) = @_;
        for my $channel (@CHANNELS) {
            $conn->privmsg($channel, $command);
        }
    }
}

{
    package Rule;

    use strict;
    use XINI;

    sub new ($$) {
        my ($class, $name) = @_;
        my $this = {};
        my ($rdh, $wrh);

        my ($section, $service);
        if ($name =~ m/^\[\s*(service (.*?))\s*\]$/) {
            $section = $1;
            $service = $2;
        } else {
            $section = "rule $name";
            $service = ini "rule $name:service";
        }
        $this->{section} = $section;
        die "No service '$service' (required for rule '$name')"
            unless exists $SERVICES{$service};

        $this->{name} = $name;
        $this->{service} = $SERVICES{$service};

        if (ini_got "$this->{section}:events") {
            $this->{events} = {map {($_ => 1)} split /\s+/, lc ini "$this->{section}:events"};
        }

        if (ini_got "$this->{section}:requests") {
            $this->{requests} = {map {($_ => 1)} split /\s+/, ini "$this->{section}:requests"};
        }

	$this->{level} = ini "$this->{section}:level"
	    if ini_got "$this->{section}:level";

        return bless $this, $class;
    }

    # Does this rule match this event?
    sub matches ($$) {
        my ($this, $event) = @_;

        # Is the event type right?
        if (exists $this->{events}) {
            #print "Checking type (is ".$event->type.", requiring @{[keys %{$this->{events}}]})\n";
            return 0 unless exists $this->{events}->{$event->type};
        }

        # Is the request prefix right?
        if (exists $this->{requests}) {
            my ($request, $args) = split /\s+/, join(' ', $event->args), 2;
            return 0 unless exists $this->{requests}->{$request};
        }

        return 1;
    }

    # Does this rule honour this event?
    # Honoured events end up having the regular action performed.
    # Non-honoured events end up getting the requester insulted.
    sub honourable ($$) {
        my ($this, $event) = @_;

	return 1 unless exists $this->{level};
	return &main::userlevel($event->userhost) >= $this->{level};
    }

    sub enter ($$) {
        my ($this, $event) = @_;

        if ($this->matches($event)) {
	    if ($this->honourable($event)) {
	        $this->{service}->handle($event);
	    } else {
                &main::insult($event);
            }
	}
        return;
    }
}

@CHANNELS = split /\s+/, ini 'general:channels = #prototux';

my $active = 0;

sub on_irc_event {
    my ($this, $event) = @_;

    $processing_event = 1;
    print $event->type, ' <', $event->nick, '> ', join(' ', $event->args), "\n";
    if ($event->type eq 'endofmotd') {
        for my $channel (@CHANNELS) {
            $this->join($channel);
        }
        $active = 1;
    }
    if ($active) {
        for my $handler (values %SERVICES, values %RULES) {
            $handler->enter($event);
        }
    }
    $processing_event = 0;
    if ($reinitialisation_pending) {
        initialise;
	$reinitialisation_pending = 0;
    }
    return;
}

$SIG{CHLD} = sub {
    while (1) {
        my $pid = waitpid -1, WNOHANG;
        last if $pid <= 0;
    }
};

$SIG{PIPE} = 'IGNORE';

$conn->add_default_handler(\&on_irc_event);

my @types = qw(nick quit join part mode topic kick public msg notice other invite kill disconnect leaving umode error cping cversion csource ctime cdcc cuserinfo cclientinfo cerrmsg cfinger caction crping crversion crsource crtime cruserinfo crclientinfo crfinger dcc_open dcc_update dcc_close chat 251 252 253 254 302 255 376 433 353);

$conn->add_global_handler(\@types, \&on_irc_event);
$conn->add_handler(\@types, \&on_irc_event);

$SIG{HUP} = sub {
    print "SIGHUP\n";
    if ($processing_event) {
        $reinitialisation_pending = 1;
    } else {
        initialise;
    }
};

$SIG{TERM} = sub {
    print "SIGTERM\n";
    $active = 0;
    $conn->quit();
    exit 0;
};

$SIG{USR1} = sub {
    print "SIGUSR1\n";
    $active = 0;
    $conn->quit();
    sleep 5;
    exec $0, @ARGV;
};

$irc->start;

sub initialise () {
    print "Using configuration from (@CONFIGFILES)\n";
    init_ini @CONFIGFILES;
    load_user_patterns 'users.pat';

    %SERVICES = ();
    for my $name (ini_sections 'service *') {
        my $service = new Service $name;
        $SERVICES{$name} = $service;
        $service->{main_rule} = new Rule "[service $name]";
    }

    %RULES = ();
    for my $name (ini_sections 'rule *') {
        my $rule = new Rule $name;
        $RULES{$name} = $rule;
    }
}

sub load_user_patterns ($) {
    my ($filename) = @_;

    @user_patterns = (); # clear out any previous patterns
    my $fh;
    open $fh, '<', $filename
        or die "$filename: open for reading: $!";
    while (<$fh>) {
        chomp;
        next if m/^#/ or m/^\s*$/;
        if (m/^\s*(\??)\s*(.+?)\@(.+?)\s*([|=])\s*(\d+?)\s*$/) {
            my ($dynflag, $usermask, $hostmask, $delim, $level) = ($1, $2, $3, $4, $5);
            if ($delim eq '=') { # glob ?
                $usermask =~ s/([^\w*])/\\$1/g; $usermask =~ s/\*/.*/g;
		unless ($dynflag) {
		    $hostmask =~ s/([^\w*])/\\$1/g; $hostmask =~ s/\*/.*/g;
		}
            }
	    $hostmask = $hostmask;
            push @user_patterns, [$dynflag, $usermask.'@'.$hostmask, $level];
        } else {
            warn "$filename:$.: unable to parse host file line \"$_\"\n"
        }
    }
    close $fh;
    return;
}

sub userlevel ($) {
    my ($userhost) = @_;

    my $result; # the default
#    if (exists $LEVEL_CACHE{$userhost}) {
#        ($result, my $timestamp) = @{$LEVEL_CACHE{$userhost}};
#	if (time - $timestamp <= 60) {
#	    return $result;
#	}
#    }
    for my $entry (@user_patterns) {
        my ($dynflag, $mask, $level) = @$entry;
	unless ($dynflag) {
            # simple, static hostmask entry
            if ($userhost =~ /^\~?$mask$/i) {
	        $result = $level;
		goto FOUND;
	    }
        } else {
            # dynamic hostmask entry
            my ($usermask, $host) = $mask =~ m/^\~?(.*?)@(.*?)$/
                or next;
            # Determine the canonical name of $host by two-way DNS resolution:
	    my $he = gethost($host);
	    next unless $he;
            for my $addr (@{$he->addr_list}) {
                my $chost = gethostbyaddr($addr)->name;
                $chost =~ s/([^\w])/\\$1/g;
                if ($userhost =~ /^\~?$usermask\@$chost$/i) {
		    $result = $level;
		    goto FOUND;
		}
            }
        }
    }
    $result = 0; # the default
FOUND:
#    $LEVEL_CACHE{$userhost} = [$result, time];
    return $result;
}

# Replaces @FOO@ in $string with $subst{FOO} or, if it's a sub ref, with
# &{$subst{FOO}}($string).  This version additionally supports the :base64
# and :bs processing flags.
sub atexpand ($%) {
    my ($string, %subst) = @_;
    my @parts = split /\@/, $string, -1;
    return '' unless @parts; # avoid OOB problems later
    my $result = $parts[0];
    my $i;
    for ($i = 1; $i < $#parts; $i++) {
        my $p = $parts[$i];
	my $suffix = '';
	$suffix = $1 if $p =~ s/(:(base64|bs))$//;
        if (exists $subst{$p}) {
            my $repl = $subst{$p};
            $repl = &$repl($p) if ref $repl eq 'CODE';
	    $repl =~ s/([^\w])/\\$1/g if $suffix =~ m/\bbs\b/i;
	    $repl = encode_base64($repl, "") if $suffix =~ m/\bbase64\b/i;
            $result .= $repl.$parts[++$i];
        } else {
            $result .= "\@$p$suffix";
        }
    }
    $result .= "\@$parts[$#parts]" if $i == $#parts;
    return $result;
}

# Takes an event and processes an autogenerated insult for it.
# This is done by referring the event to the 'autoinsult' service.
# If there is no such service, the user is left uninsulted.
sub insult ($) {
    my ($event) = @_;

    print "autoinsulting ", $event->nick, "\n";
    $SERVICES{autoinsult}->handle($event)
        if exists $SERVICES{autoinsult};
}
